"""unicloud.

Python package doing xyz for .com

https://gitlab.com/cquest1/tools/unicloud
"""
try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

__author__ = "Mostafa Farrag"
__email__ = "mostafa.farrag@.com"
