"""unicloud.

Python package doing xyz for .com

https://gitlab.com/cquest1/tools/unicloud
"""


def main():
    """Execute unicloud functionalities."""
    return True


if __name__ == "__main__":
    main()
